from typing import List
from langchain.pydantic_v1 import BaseModel, Field


class Theme(BaseModel):
    """ "A recurring idea or concept in a work of literature"""

    name: str = Field(
        ..., description="The name of the theme. Ex: Death and Mortality, Regret, etc."
    )
    reasoning: str = Field(
        ...,
        description="The reasoning behind the theme occuring in a specific peice of text",
    )


class Character(BaseModel):
    """A person or being in a work of literature"""

    name: str = Field(..., description="The name of the character")
    description: str = Field(
        ..., description="A description of the character and their role in the story"
    )
    physical_description: str = Field(
        ..., description="A description of the character's physical appearance"
    )
    known_as: List[str] = Field(
        ..., description="A list of names the character is known as in the story"
    )
    summary_of_actions: str = Field(
        ..., description="A summary of the actions the character takes in the story"
    )



from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import cached_property
from typing import List, Sequence

from langchain.prompts import ChatPromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.chains.openai_functions import (
    create_structured_output_runnable,
)
from langchain.pydantic_v1 import BaseModel, Field, create_model


class Consolidator:
    """
    A class that consolidates a list of target objects into a grouped list.

    Args:
        target_class (type[BaseModel]): The class of the target objects.
        target_num (int, optional): The desired number of target objects in the grouped list.
            Defaults to 3.
        model (str, optional): The model to use for consolidation. Defaults to "gpt-4".
        batch_size (int, optional): The batch size for processing the target objects.
            Defaults to 10.
    """

    def __init__(
        self,
        target_class: type[BaseModel],
        target_num: int = 3,
        model: str = "gpt-4",
        batch_size: int = 10,
        max_iter: int = 3,
    ):
        # Vars
        self.target_name = target_class.__name__
        self.target_desc = target_class.__doc__
        self.target_num = target_num
        self.llm = ChatOpenAI(model=model)
        self.batch_size = batch_size
        self.max_iter = max_iter

        # Grouped class
        fields = {
            f"{self.target_name}s": (
                Sequence[target_class],
                Field(..., description=f"A list of {self.target_name}s"),
            ),
        }
        grouped_type = create_model(f"{self.target_name}s", **fields)
        grouped_type.__doc__ = f"A list of {self.target_name}s"
        self.grouped_target_type = grouped_type

        # objects
        self.consolidation_chain = self._create_consolidation_chain()

    @cached_property
    def _consolidation_prompt(self) -> ChatPromptTemplate:
        return ChatPromptTemplate.from_template(
            f"Please consolidate the following {self.target_name}s into a list: \n"
            + "{targets}"
            + f"\n Consolidate into {self.target_num} {self.target_name}s."
            + f"Where possible, combine similar {self.target_name}s into one."
            + f"Look for cases where the same {self.target_name} is referred to by "
            + "different names. Ex: The Woman and Helen."
        )

    def _create_consolidation_chain(self):
        runnable = create_structured_output_runnable(
            output_schema=self.grouped_target_type,
            llm=self.llm,
            prompt=self._consolidation_prompt,
        )
        return runnable

    def consolidate(self, items: List[BaseModel], current_iter: int = 0) -> list:
        """
        Consolidates a list of target objects into a grouped list.

        Args:
            items (List[BaseModel]): The list of target objects to be consolidated.

        Returns:
            list: The consolidated grouped list of target objects.
        """
        batches = [
            items[i : i + self.batch_size]
            for i in range(0, len(items), self.batch_size)
        ]

        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [executor.submit(self._consolidate, batch) for batch in batches]
            results = []
            for future in as_completed(futures):
                result = future.result()
                results.extend(result)

        if len(results) >= self.target_num * 1.5 and current_iter < self.max_iter:
            return self.consolidate(results, current_iter + 1)

        return results

    def _consolidate(self, results: List[BaseModel]) -> list:
        """
        Consolidates a batch of target objects using the consolidation chain.

        Args:
            results (List[BaseModel]): The batch of target objects to be consolidated.

        Returns:
            list: The consolidated list of target objects.
        """
        # consolidate with agent
        consolidated_results = self.consolidation_chain.invoke({"targets": results})
        result = getattr(consolidated_results, self.target_name + "s")
        return result



from typing import Generator, Iterable


def text_generator(
    texts: Iterable[str], max_chunk_size: int
) -> Generator[str, None, None]:
    """
    Given a n iterable of strings, generates a stream that concatenates the strings
    into chunks of a given size.

    """
    next_chunk = ""
    for text in texts:
        # Split large text
        if len(text) > max_chunk_size:
            for i in range(0, len(text), max_chunk_size):
                yield text[i : i + max_chunk_size]

        # Return when chunk is full
        elif len(next_chunk) + len(text) > max_chunk_size:
            if next_chunk:
                yield next_chunk
            next_chunk = text
        else:
            next_chunk += text
    if next_chunk:
        yield next_chunk



from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import cached_property
import logging
from typing import Generator, Iterable, Sequence, Union

from langchain.prompts import ChatPromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.chains.openai_functions import (
    create_structured_output_runnable,
)
from langchain.pydantic_v1 import BaseModel, Field, create_model


class Extractor:
    """
    A class for extracting information from a given text using a language chain.

    Attributes:
        target_class (type[BaseModel]): The class of the target to extract from the text.
        target_num (int): The expected number of targets to extract from the text.
        target_examples (list): A list of examples of the target to extract from the text.

    Methods:
        extract(text): Extracts information from the given text using the language chain.
    """

    def __init__(
        self,
        target_class: type[BaseModel],
        target_num: int = 3,
        target_examples: list = None,
        model: str = "gpt-4",
    ):
        # Logger
        self._logger = logging.getLogger(__name__)

        # Vars
        self.target_name = target_class.__name__
        self.target_desc = target_class.__doc__
        self.target_num = target_num
        self.target_examples = target_examples or []
        self.llm = ChatOpenAI(model=model)

        # Grouped class
        fields = {
            f"{self.target_name}s": (
                Sequence[target_class],
                Field(..., description=f"A list of {self.target_name}s"),
            ),
        }
        grouped_type = create_model(f"{self.target_name}s", **fields)
        grouped_type.__doc__ = f"A list of {self.target_name}s"
        self.grouped_target_type = grouped_type

    @cached_property
    def _extraction_prompt(self) -> ChatPromptTemplate:
        prompt = (
            f"You are a literary analyst. Your job is to determine any {self.target_name}s "
            "present in a given text."
            f"{self.target_name} is defined as {self.target_desc}"
            "Here is the text: \n {text}"
            f"\n Please determine any {self.target_name}s present in the text. "
            f"Look for around {{target_num}} {self.target_name}s."
        )
        if self.target_examples:
            prompt += f"\n ex: {self.target_examples}"

        return ChatPromptTemplate.from_template(prompt)

    def _create_extraction_chain(self):
        runnable = create_structured_output_runnable(
            output_schema=self.grouped_target_type,
            llm=self.llm,
            prompt=self._extraction_prompt,
        )
        return runnable

    def _extract_from_block(self, text: str, number: int = None) -> list:
        """
        Extracts information from the given text using the language chain.

        Args:
            text (str): The text to extract information from.

        Returns:
            list: A list of extracted information from the given text.
        """
        extraction_chain = self._create_extraction_chain()
        result = extraction_chain.invoke(
            {"text": text, "target_num": number or self.target_num}
        )
        return getattr(result, self.target_name + "s")

    def extract(
        self, input_data: Union[str, Iterable, Generator], chunk_size=4000, **kwargs
    ) -> list:
        pool = ThreadPoolExecutor(max_workers=10)
        futures = []

        if isinstance(input_data, str):
            # Input is a string, process it in chunks
            for i in range(0, len(input_data), chunk_size):
                text_chunk = input_data[i : i + chunk_size]
                futures.append(
                    pool.submit(self._extract_from_block, text_chunk, **kwargs)
                )
        else:
            # Input is a stream or generator, process it iteratively
            text_chunk = ""
            for piece in input_data:
                text_chunk += piece
                if len(text_chunk) >= chunk_size:
                    # Submit the current chunk for processing
                    futures.append(
                        pool.submit(self._extract_from_block, text_chunk, **kwargs)
                    )
                    text_chunk = ""
            # Make sure to process the last chunk if it's not empty
            if text_chunk:
                futures.append(
                    pool.submit(self._extract_from_block, text_chunk, **kwargs)
                )

        # Retrieve results as they complete
        results = []
        completed = 0
        for future in as_completed(futures):
            # Get the result of the future
            result = future.result()
            # Since the result itself might be a list, we extend our result list with it
            results.extend(result)
            completed += 1
            self._logger.debug(f"Finished {completed} of {len(futures)}")

        return results



from concurrent.futures import ThreadPoolExecutor
from functools import cached_property
import logging
from typing import List


from langchain.prompts import ChatPromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.chains.openai_functions import (
    create_structured_output_runnable,
)
from langchain.pydantic_v1 import BaseModel


class Annotator:
    def __init__(
        self,
        targets=List[BaseModel],
        model: str = "gpt-4",
    ):
        # logger
        self._logger = logging.getLogger(__name__)

        # Vars
        target_class = targets[0].__class__
        self.target_name = target_class.__name__
        self.target_desc = target_class.__doc__
        self.targets = targets
        self.llm = ChatOpenAI(model=model)

        # Schema
        self.json_schema = {
            "title": f"{self.target_name}s",
            "description": (
                f"List of {self.target_name}s indices that are present in the text, 0-based. For example, "  # noqa: E501
                f"if {self.targets[0]} is present in the text, then include 0 in the response."
            ),
            "type": "object",
            "properties": {
                "indices": {
                    "title": f"{self.target_name}s",
                    "description": f"List of 0-based indices of {self.target_name}s that are present in the text.",  # noqa: E501
                    "type": "array",
                    "items": {
                        "type": "number",
                        "enum": [i for i in range(len(self.targets))],
                    },
                },
            },
        }

    @cached_property
    def _annotation_prompt(self) -> ChatPromptTemplate:
        return ChatPromptTemplate.from_template(
            f"You are tasked with annotating a text sample to identify {self.target_name}s. "
            f"Here is an ordered list of valid targets: \n {self.targets}\n"
            "Here is the text sample: \n {text} \n"
            f"Please return all of the given {self.target_name} indices that are in the text."
        )

    def _create_annotation_chain(self):
        runnable = create_structured_output_runnable(
            output_schema=self.json_schema,
            llm=self.llm,
            prompt=self._annotation_prompt,
        )
        return runnable

    def _annotate(self, text: str) -> list:
        extraction_chain = self._create_annotation_chain()
        result = extraction_chain.invoke({"text": text})
        return result["indices"]

    def annotate(self, text: str, chunk_size=4000, **kwargs) -> set:
        pool = ThreadPoolExecutor(max_workers=10)

        # Split text into chunks and extract asynchronously
        futures = []
        for i in range(0, len(text), chunk_size):
            text_chunk = text[i : i + chunk_size]
            futures.append(pool.submit(self._annotate, text_chunk, **kwargs))

        # Retrieve results in order
        results = []
        completed = 0
        for future in futures:
            # Get the result of the future
            result = future.result()
            results.extend(result)
            completed += 1
            self._logger.debug(f"Finished {completed} of {len(futures)}")

        return set(results)



from abc import ABC, abstractmethod
from typing import Optional


class Segmentor(ABC):
    """
    Class for segmenting text into various organizational units, such as chapters, paragraphs,
    sentences, scenes, etc.
    """

    def __init__(self, max_length: Optional[int] = None, handle_length: str = "split") -> None:
        self._max_length = max_length
        allowed_handle_lengths = ["split", "truncate", "raise"]
        if handle_length not in allowed_handle_lengths:
            raise ValueError("handle_length must be one of {}".format(allowed_handle_lengths))
        self._handle_length = handle_length

    @abstractmethod
    def _segment(self, text: str) -> list[str]:
        """
        Segments the given text into organizational units. Override this for
        specific implementations
        """
        pass  # pragma: no cover

    def segment(self, text: str) -> list[str]:
        """
        Segments the given text into organizational units.
        """
        text_segments = self._segment(text)
        return self._handle_max_length(text_segments)

    def _handle_max_length(self, text_segments: list[str]) -> list[str]:
        """
        Splits the text into segments of the maximum length.
        """
        if self._max_length is None:
            return text_segments

        segments = []
        for i, segment in enumerate(text_segments):
            if len(segment) > self._max_length:
                if self._handle_length == "split":
                    segments.extend(
                        (segment[0: self._max_length], segment[self._max_length:])
                    )
                elif self._handle_length == "truncate":
                    segments.append(segment[: self._max_length])
                elif self._handle_length == "raise":
                    raise ValueError(
                        f"Segment {i} length {len(segment)} "
                        f"exceeds maximum length {self._max_length}"
                    )
            else:
                segments.append(segment)

        return segments


class ParagraphSegmentor(Segmentor):
    """
    A class that segments text into paragraphs.
    """

    def _segment(self, text: str) -> list:
        """
        Segments the given text into paragraphs.
        """
        return [p for p in text.split("\n") if p]


class PageSegmentor(Segmentor):
    """
    A class that segments text into pages. Each page contains 10 paragraphs.
    """

    def __init__(self, paragraphs_per_page: int = 10, **kwargs) -> None:
        super().__init__(**kwargs)
        self._paragraphs_per_page = paragraphs_per_page

    def _segment(self, text: str) -> list:
        """
        Segments the given text into pages.
        """
        paragraphs = [p for p in text.split("\n") if p]
        return [
            "\n".join(paragraphs[i : i + self._paragraphs_per_page])
            for i in range(0, len(paragraphs), self._paragraphs_per_page)
        ]



from .annotator import Annotator
from .consolidator import Consolidator
from .extractor import Extractor
from .processor import Processor
from .segmentor import ParagraphSegmentor, PageSegmentor

__all__ = [
    "Annotator",
    "Consolidator",
    "Extractor",
    "Processor",
    "ParagraphSegmentor",
    "PageSegmentor"
]



import logging

from langchain.pydantic_v1 import BaseModel

from .consolidator import Consolidator
from .extractor import Extractor
from .. import utils


class TextmancyResult(BaseModel):
    """
    A class that represents the result of processing a text corpus.
    """

    targets: list
    annotations: list[int]


class Processor:
    def __init__(
        self,
        target_class: type[BaseModel],
        model: str = "gpt-4-1106-preview",
        target_num_hint: int = 10,
    ):
        self.extractor = Extractor(
            target_class=target_class, model=model, target_num=target_num_hint
        )
        self.consolidator = Consolidator(target_class=target_class, model=model)
        self._logger = logging.getLogger(__name__)

    def process(self, texts: list[str]) -> TextmancyResult:
        """
        Extracts featres from text, consolidates and then annotates the given text fragments.
        """

        # Create text stream for extractor
        self._logger.info("Creating text stream for extraction")
        text_stream = utils.text_generator(texts, max_chunk_size=10000)

        # Extract and consolidate
        self._logger.info("Extracting features")
        results = self.extractor.extract(text_stream)
        self._logger.debug(f"Extracted {len(results)} features")

        self._logger.info("Consolidating features")
        consolidated = self.consolidator.consolidate(results)
        self._logger.debug(f"Consolidated into {len(consolidated)} features")

        return consolidated
